use crate::app_config::AppConfig;
use asn1::{
    parse_single, Asn1Readable, Asn1Writable, Implicit, OctetStringEncoded, ParseResult,
    SimpleAsn1Readable, SimpleAsn1Writable, Tag, WriteBuf, WriteResult, Writer,
};
use openssl::asn1::{Asn1Object, Asn1OctetString, Asn1OctetStringRef, Asn1Time};
use openssl::bn::{BigNum, MsbOption};
use openssl::ec::EcGroup;
use openssl::hash::MessageDigest;
use openssl::nid::Nid;
use openssl::pkey::{PKey, Private};
use openssl::x509;
use openssl::x509::extension::{BasicConstraints, KeyUsage, SubjectKeyIdentifier};
use openssl::x509::{X509Extension, X509NameBuilder, X509};
use std::path::Path;

/// Certificate and private key
struct CertAndKey(X509, PKey<Private>);

impl CertAndKey {
    /// Load root CA
    fn load_root_ca() -> anyhow::Result<Self> {
        Ok(Self(
            load_certificate_from_file(AppConfig::get().root_ca_cert_path())?,
            load_priv_key_from_file(AppConfig::get().root_ca_priv_key_path())?,
        ))
    }
}

/// Generate private key
fn gen_private_key() -> anyhow::Result<PKey<Private>> {
    let nid = Nid::X9_62_PRIME256V1; // NIST P-256 curve
    let group = EcGroup::from_curve_name(nid)?;
    let key = openssl::ec::EcKey::generate(&group)?;
    let key_pair = PKey::from_ec_key(key.clone())?;

    Ok(key_pair)
}

/// Load private key from PEM file
fn load_priv_key_from_file<P: AsRef<Path>>(path: P) -> anyhow::Result<PKey<Private>> {
    Ok(PKey::private_key_from_pem(&std::fs::read(path)?)?)
}

/// Load certificate from PEM file
fn load_certificate_from_file<P: AsRef<Path>>(path: P) -> anyhow::Result<X509> {
    Ok(X509::from_pem(&std::fs::read(path)?)?)
}

struct CustomOctetStringEncoded<T>(OctetStringEncoded<T>);
impl<T: Asn1Writable> SimpleAsn1Writable for CustomOctetStringEncoded<T> {
    const TAG: Tag = Tag::primitive(0x86);
    fn write_data(&self, dest: &mut WriteBuf) -> WriteResult {
        self.0.write(&mut Writer::new(dest))
    }
}

/// Generate intermediate or root CA
fn gen_intermediate_or_root_ca(
    cn: &str,
    issuer: Option<&CertAndKey>,
) -> anyhow::Result<(Vec<u8>, Vec<u8>)> {
    // Generate root private key
    let key_pair = gen_private_key()?;

    let mut x509_name = X509NameBuilder::new()?;
    x509_name.append_entry_by_text("C", "FR")?;
    x509_name.append_entry_by_text("CN", cn)?;
    let x509_name = x509_name.build();

    let mut cert_builder = X509::builder()?;
    cert_builder.set_version(2)?;
    let serial_number = {
        let mut serial = BigNum::new()?;
        serial.rand(159, MsbOption::MAYBE_ZERO, false)?;
        serial.to_asn1_integer()?
    };
    cert_builder.set_serial_number(&serial_number)?;
    cert_builder.set_subject_name(&x509_name)?;
    match issuer {
        // Self-signed certificate
        None => cert_builder.set_issuer_name(&x509_name)?,
        // Certificate signed by another CA
        Some(i) => cert_builder.set_issuer_name(i.0.issuer_name())?,
    }
    cert_builder.set_pubkey(&key_pair)?;
    let not_before = Asn1Time::days_from_now(0)?;
    cert_builder.set_not_before(&not_before)?;
    let not_after = Asn1Time::days_from_now(365 * 30)?;
    cert_builder.set_not_after(&not_after)?;

    if let Some(issuer) = issuer {
        let crl_url = format!(
            "{}/crl/{}.crl",
            AppConfig::get().unsecure_origin(),
            "FIXME_TODO"
        );

        let crl_obj = Asn1Object::from_str("2.5.29.31")?;

        let content: Implicit<CustomOctetStringEncoded<&[u8]>, 0xa0> = asn1::Implicit::new(
            CustomOctetStringEncoded(OctetStringEncoded::new(crl_url.as_bytes())),
        );

        let crl_bytes = asn1::write(|w| {
            w.write_element(&asn1::SequenceWriter::new(&|w| {
                w.write_element(&asn1::SequenceWriter::new(&|w| {
                    w.write_implicit_element(&content, 0xa0)?;
                    Ok(())
                }))?;
                Ok(())
            }))
        })?;

        cert_builder.append_extension(X509Extension::new_from_der(
            crl_obj.as_ref(),
            false,
            Asn1OctetString::new_from_bytes(&crl_bytes)?.as_ref(),
        )?)?;
    }

    cert_builder.append_extension(BasicConstraints::new().critical().ca().build()?)?;

    cert_builder.append_extension(
        KeyUsage::new()
            .critical()
            .key_cert_sign()
            .crl_sign()
            .build()?,
    )?;

    let subject_key_identifier =
        SubjectKeyIdentifier::new().build(&cert_builder.x509v3_context(None, None))?;

    cert_builder.append_extension(subject_key_identifier)?;

    cert_builder.sign(
        match issuer {
            None => &key_pair,
            Some(i) => &i.1,
        },
        MessageDigest::sha256(),
    )?;
    let cert = cert_builder.build();

    Ok((key_pair.private_key_to_pem_pkcs8()?, cert.to_pem()?))
}

/// Initialize Root CA, if required
pub fn initialize_root_ca() -> anyhow::Result<()> {
    if AppConfig::get().root_ca_cert_path().exists()
        && AppConfig::get().root_ca_priv_key_path().exists()
    {
        return Ok(());
    }

    log::info!("Generating root ca...");

    let (key, cert) = gen_intermediate_or_root_ca("SolarEnergy Root CA", None)?;

    // Serialize generated web CA
    std::fs::write(AppConfig::get().root_ca_priv_key_path(), key)?;
    std::fs::write(AppConfig::get().root_ca_cert_path(), cert)?;

    Ok(())
}

/// Initialize web CA, if required
pub fn initialize_web_ca() -> anyhow::Result<()> {
    if AppConfig::get().web_ca_cert_path().exists()
        && AppConfig::get().web_ca_priv_key_path().exists()
    {
        return Ok(());
    }

    log::info!("Generating web ca...");

    let (key, cert) =
        gen_intermediate_or_root_ca("SolarEnergy Web CA", Some(&CertAndKey::load_root_ca()?))?;

    // Serialize generated web CA
    std::fs::write(AppConfig::get().web_ca_priv_key_path(), key)?;
    std::fs::write(AppConfig::get().web_ca_cert_path(), cert)?;

    Ok(())
}

/// Initialize devices CA, if required
pub fn initialize_devices_ca() -> anyhow::Result<()> {
    if AppConfig::get().devices_ca_cert_path().exists()
        && AppConfig::get().devices_ca_priv_key_path().exists()
    {
        return Ok(());
    }

    log::info!("Generating devices ca...");

    let (key, cert) =
        gen_intermediate_or_root_ca("SolarEnergy Devices CA", Some(&CertAndKey::load_root_ca()?))?;

    // Serialize generated devices CA
    std::fs::write(AppConfig::get().devices_ca_priv_key_path(), key)?;
    std::fs::write(AppConfig::get().devices_ca_cert_path(), cert)?;

    Ok(())
}
